drop1.lm <-
function(object, scope, scale, keep, all.cols = T)
{
	b <- coef(object)
	cnames <- labels(b)
	singular <- attr(b, "singular")
	p <- length(b)
	x <- model.matrix(object)
	iswt <- !is.null(wt <- object$weights)
	if(iswt)
		x <- x * sqrt(wt)
	n <- dim(x)[[1]]
	asgn <- attr(x, "assign")
	tl <- attr(object$terms, "term.labels")
	if(missing(scope))
		scope <- drop.scope(object)
	else {
		if(!is.character(scope))
			scope <- attr(terms(update.formula(object, scope)), "term.labels")
		if(!all(match(scope, tl, F)))
			stop("scope is not a subset of term labels")
	}
	asgn <- asgn[scope]
	k <- length(scope)
	rdf <- object$df.resid
	chisq <- deviance.lm(object)
	if(missing(scale))
		scale <- chisq/rdf
	if(!missing(keep)) {
		max.keep <- c("coefficients", "fitted", "residuals", "x.residuals", "effects", 
			"R")
		if(is.logical(keep) && keep)
			keep <- max.keep
		else {
			if(!all(match(keep, max.keep, F)))
				stop(paste("Can only keep one or more of: \"", paste(max.keep, 
				  collapse = "\", \""), "\"", sep = ""))
		}
		fv <- predict(object)
		y <- object$residuals + fv
		if(iswt) {
			if(any(wt == 0))
				stop("\"keep\" not allowed when some of the weights are zero")
			wt <- sqrt(wt)
		}
	}
	else keep <- character(0)
	xr <- match("x.residuals", keep, F)
	value <- array(vector("list", 6 * k), c(k, 6), list(scope, c("coefficients", "fitted", 
		"residuals", "x.residuals", "effects", "R")))
	if(length(ef <- object$effects) < n)
		stop("function only currently defined for methods that compute effects")
	dfs <- double(k)
	chis <- double(k)
	if(singular) {
		y <- (object$residuals + predict(object))
		na.coef <- (1:length(object$coefficients))[!is.na(object$coefficients)]
		if(iswt) {
			if(!length(keep))
				wt <- sqrt(wt)
			y <- y * wt
		}
		rank <- object$rank
		for(i in 1:k) {
			ii <- asgn[[i]]	#brute force method
			if(all.cols)
				jj <- setdiff(seq(ncol(x)), ii)
			else 
				jj <- setdiff(na.coef, ii)
			z <- lm.fit.qr(x[, jj, drop = F], y, singular = T, qr = xr)
			efi <- z$effects
			dfs[i] <- rank - (ranki <- z$rank)
			chis[i] <- sum(efi[ - seq(ranki)]^2)
			if(length(keep)) {
				fvi <- z$fitted
				res <- z$residuals
				if(iswt) {
				  fvi <- fvi/wt
				  res <- res/wt
				}
				value[i, -4] <- list(z$coef, fvi, res, efi, z$R)
				if(xr) {
				  xres <- qr.resid(z$qr, x[, ii, drop = F])
				  if(iswt)
				    xres <- xres/wt
				  value[[i, 4]] <- xres
				}
			}
		}
	}
	else {
		R <- object$R
		R <- array(R, dim(R), dimnames(R))
		if(xr) {
			xk <- array(0, dim(x), dimnames(x))
			xk[1:p,  ] <- R
		}
		else xk <- array(ef, c(n, 1), list(dimnames(x)[[1]], NULL))
		for(i in 1:k) {
			ii <- asgn[[i]]
			pii <- length(ii)
			dfs[i] <- pii
			if(xr) {
				xi <- xk[, c(1, ii)]
				xi[, 1] <- ef
			}
			else xi <- xk
			r <- R
			pi <- 1:(p - pii)
			pp <-  - p
			for(j in rev(ii)) {
				z <- delcol(r, xi, j)
				r <- z[[1]][pp,  ]
				xi <- z[[2]]
				pp <- pp + 1
			}
			efi <- xi[, 1]
			chis[i] <- sum(efi[ - pi]^2)
			if(length(keep)) {
# compute it all, even though all may not be reqd
				bi <- as.matrix(backsolve(r, xi[pi,  ]))
				dimnames(bi)[[1]] <- cnames[ - ii]
				fvi <- x[,  - ii, drop = F] %*% bi
				if(iswt)
				  fvi <- fvi/wt
				names(efi)[] <- ""
				names(efi)[pi] <- cnames[ - ii]
				value[i, -4] <- list(bi[, 1], fvi[, 1], y - fvi[, 1], efi, r)
				if(xr) {
				  xres <- x[, ii] - fvi[, -1]
				  if(iswt)
				    xres <- xres/wt
				  value[[i, 4]] <- xres
				}
			}
		}
	}
	scope <- c("<none>", scope)
	dfs <- c(0, dfs)
	chis <- c(chisq, chis)
	aics <- chis + 2 * (n - rdf - dfs) * scale
	dfs[1] <- NA
	aod <- data.frame(Df = dfs, "Sum of Sq" = c(NA, chis[-1] - chis[1]), RSS = chis, Cp = 
		aics, row.names = scope, check.names = F)
	head <- c("Single term deletions", "\nModel:", deparse(as.vector(formula(object))))
	if(!missing(scale))
		head <- c(head, paste("\nscale: ", format(scale), "\n"))
	class(aod) <- c("anova", "data.frame")
	attr(aod, "heading") <- head
	if(length(keep))
		list(anova = aod, keep = structure(value[, keep, drop = F], class = "matrix"))
	else aod
}
